﻿using System;
using System.Data;
using System.Configuration;
using System.Collections;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Web.UI.HtmlControls;
using System.Collections.Generic;
using System.Threading;
using DSearchCommon;
using Local;
using System.Runtime.Remoting.Channels.Tcp;
using System.Runtime.Remoting.Channels;
namespace DSearchWeb
{
    /// <summary>
    /// 状态对象
    /// </summary>
    public class StateObj
    {
        public StateObj( ISearcher searcher,
            string keyWord,
            AutoResetEvent are)
        {
            Searcher = searcher;
            KeyWord = keyWord;
            AutoEvent = are;
        }
        public AutoResetEvent AutoEvent;
        public string KeyWord;
        public ISearcher Searcher;
    }

    public partial class _Default : System.Web.UI.Page
    {
        protected void Page_Load(object sender, EventArgs e)
        {

        }
        List<SearchResult> m_list = new List<SearchResult>();   //全局搜索结果列表
        Int32 numAsyncOps = 2;                                  //异步操作的数量
        AutoResetEvent areDoSearch = new AutoResetEvent(false); //全部搜索完成的事件
        AutoResetEvent areDoFilter = new AutoResetEvent(false); //敏感此过滤完成的事件

        //处理搜索按钮的方法
        protected void btnSearch_Click(object sender, EventArgs e)
        {
            getSearchResults(txtKeyword.Text.Trim());            
            rptResults.DataSource = m_list;
            rptResults.DataBind();
        }

        /// <summary>
        /// 进行多线程和分布式搜索
        /// </summary>
        /// <param name="keyWord"></param>
        private void getSearchResults(string keyWord)
        {
            //注册全部搜索完成后将执行的回调，回调里将结果进行敏感词过滤
            //这里要设置一个合理的超时值，以防止有些服务器操作超时造成整体无响应
            RegisteredWaitHandle rwh = ThreadPool.RegisterWaitForSingleObject(
                areDoSearch,
                doFiterResult,
                null,
                10000,
                false);

            #region 创建搜索服务器列表
            ISearcher[] arrSearcher;
            try
            {
                //这里用一个子方法来完成所有SearchServer的创建
                //实际应用中应该用单件模式来实现，而不是每次搜索
                //都创建一遍
                arrSearcher = getSearcherList();

            }
            catch (Exception ex)
            {
                Trace.Warn("访问搜索服务器出错：" + ex.ToString());
                return;
            } 
            #endregion

            numAsyncOps = arrSearcher.Length; //重置异步操作数量

            #region 调用线程池线程分别执行每个远程异步搜索
            AutoResetEvent[] ares = new AutoResetEvent[arrSearcher.Length];
            int iSearcher = 0;
            StateObj stateObj = null;
            foreach (ISearcher searcher in arrSearcher)
            {
                //每个异步操作拥有一个自动事件，用来通知主线程完成还是超时
                ares[iSearcher] = new AutoResetEvent(false); 
                stateObj = new StateObj(searcher, keyWord, ares[iSearcher]); //准备状态参数
                
                try
                {
                    if (!ThreadPool.QueueUserWorkItem(new WaitCallback(doSearch),
                        stateObj))
                    {
                        //如果无法压入线程池工作队列，记录下日志
                        Trace.Warn("Unable to queue ThreadPool request.");
                    }
                }
                catch (ApplicationException ae) //调用QueueUserWorkItem有可能会有异常
                {
                    Trace.Warn("服务器内存不足" + ae.ToString());
                }
                iSearcher++;
            } 
            #endregion

            #region 等待每一个异步搜索直到完成或者超时
            //这个等待主要是为了防止某台服务器超时，否则
            //本循环的最后两句可以写到doSearch的最后
            //但是那样如果有一台服务器长时间无响应，这个搜索过程就会无响应了
            foreach (AutoResetEvent are in ares)
            {
                if (are.WaitOne(new TimeSpan(0, 0, 15), false))
                {
                    //这里只记录某台，实际中应该记录详细的机器编号
                    Trace.Warn("某台搜索服务器执行成功");
                }
                else
                {
                    Trace.Warn("某台搜索服务器执行超时");
                }
                //如果全部异步搜索已完成或超时，那么设置areDoSearch为阻止状态
                //此时会触发doFiterResult方法去对结果进行敏感词过滤
                //Interlocked保证递增和递减为一个CPU的原子操作
                if (Interlocked.Decrement(ref numAsyncOps) == 0)
                    areDoSearch.Set();
            } 
            #endregion

            areDoFilter.WaitOne();  //等待敏感词过滤完毕
            rwh.Unregister(null);   //注销areDoSearch事件
        }

        /// <summary>
        /// 创建搜索服务器列表，真实情况下建议使用工厂方法模式
        /// </summary>
        /// <returns></returns>
        private ISearcher[] getSearcherList()
        {
            List<ISearcher> list= new List<ISearcher>();
            list.Add(new LocalSearcher());
            
            ISearcher searcher = (ISearcher)Activator.GetObject(typeof(ISearcher),
                "tcp://localhost:8080/RemWebSearcher");
            list.Add(searcher);
            return list.ToArray();
        }

        /// <summary>
        /// 进行结果过滤操作，真实情况下需要进行关联度分析后排序，以及敏感词过滤等操作
        /// </summary>
        /// <param name="state"></param>
        /// <param name="timeOut"></param>
        void doFiterResult(Object state, bool timeOut)
        {
            Trace.Warn("{" + Thread.CurrentThread.GetHashCode().ToString() + "}:doFiterResult...");
            if (timeOut)
            {
                Trace.Warn("搜索超时，某些搜索服务器可能没有在指定的时间内返回结果");
            }
            lock (((ICollection)m_list).SyncRoot)
            {
                //这里模拟一些敏感词过滤的逻辑
                m_list.Add(new SearchResult("嘿嘿一定是你吧"));
            }
            areDoFilter.Set();  //设置敏感词过滤事件为终止，通知主线程搜索结果已经可用，可以展示给用户
        }

        /// <summary>
        /// 每个搜索器的执行方法
        /// </summary>
        /// <param name="state"></param>
        void doSearch(Object state)
        {
            Trace.Warn("{" + Thread.CurrentThread.GetHashCode().ToString() + "}:doSearch...");

            #region 处理该线程上的未处理异常，虽然对搜索过程加了cathc，但还是要防止漏网之鱼
            AppDomain.CurrentDomain.UnhandledException += delegate(object sender,
                UnhandledExceptionEventArgs e)
                {
                    Trace.Warn(Thread.CurrentThread.GetHashCode().ToString()
                        + "线程上有未处理的异常:" +
                        ((Exception)e.ExceptionObject).ToString());
                };
            #endregion


            StateObj stateObj = (StateObj)state;

            ISearcher searcher = stateObj.Searcher; //当前的搜索器
            string keyWord = stateObj.KeyWord;      //搜索的关键词

            #region 执行搜索器的搜索方法获取搜索结果，并捕获异常
            IList<SearchResult> lst = null;
            try
            {
                lst = searcher.Search(keyWord);
            }
            catch (Exception ex)
            {
                Trace.Warn("doSearch has an error:" + ex.ToString());
            } 
            #endregion


            #region 拿到结果后，把结果合并到全局搜索结果列表里
            //考虑到多线程访问全局结果列表，这里要对其加锁
            if (null != lst)
            {
                lock (((ICollection)m_list).SyncRoot)
                {
                    m_list.AddRange(lst);
                }
            } 
            #endregion

            stateObj.AutoEvent.Set();  //通知主线程，本搜索器执行完毕，可以等下一个了。
        }

    }
}
